/******* fermion_force_hisq_qop.c ****************/
/* MIMD version 7 */

/* This is the MILC wrapper for the SciDAC Level 3 QOP fermion force routine */
/* 5/25/09  C. DeTar and A Bazavov */

/* External entry points in this file

   eo_fermion_force_oneterm
   eo_fermion_force_twoterms
   eo_fermion_force_multi
   ks_multiff_opt_chr

 */

/* Compile with fermion_force_hisq_qop_[FD].c */

/*
 * $Log: fermion_force_hisq_qop.c,v $
 */

#include "generic_ks_includes.h"
#include "../include/generic_qop.h"
#include "../include/generic_ks_qop.h"

/* Set default if undeclared */
#ifndef KS_MULTIFF
#define KS_MULTIFF FNMAT
#endif

static char* cvsHeader = "$Header: /u/aquila/cvsroot/milc_qcd/generic_ks/fermion_force_hisq_qop.c,v 1.27 2007/12/14 04:36:31 detar Exp $";

/**********************************************************************/
/* Standard MILC interface for the single-species HISQ fermion force
   routine */
/**********************************************************************/
void eo_fermion_force_oneterm( Real eps, Real weight, field_offset x_off,
			       int prec, ferm_links_t *fn, ks_action_paths *ap)
{

  if(prec == 1)
    eo_fermion_force_oneterm_F( eps, weight, x_off, fn, ap );
  else
    eo_fermion_force_oneterm_D( eps, weight, x_off, fn, ap );

}

/**********************************************************************/
/* Standard MILC interface for the two-species HISQ fermion force
   routine */
/**********************************************************************/
void eo_fermion_force_twoterms( Real eps, Real weight1, Real weight2, 
				field_offset x1_off, field_offset x2_off,
				int prec, ferm_links_t *fn, 
				ks_action_paths *ap) 
{

  if(prec == 1)
    eo_fermion_force_twoterms_F( eps, weight1, weight2, x1_off, x2_off, 
				 fn, ap );
  else
    eo_fermion_force_twoterms_D( eps, weight1, weight2, x1_off, x2_off,
				 fn, ap );

}


/**********************************************************************/
/*   Parallel transport nterms source vectors                        */
/**********************************************************************/

static void 
fermion_force_hisq_multi( Real eps, Real *residues, 
			    su3_vector **xxx, int nterms, int prec,
			    ferm_links_t *fn, ks_action_paths *ap ) 
{

  if(prec == 1)
    fermion_force_hisq_multi_F( eps, residues, xxx, nterms, fn, ap );
  else
    fermion_force_hisq_multi_D( eps, residues, xxx, nterms, fn, ap );

}

/**********************************************************************/
/*   Parallel transport vectors in blocks of veclength.               */
/**********************************************************************/
/* Requires the xxx1 and xxx2 terms in the site structure */

static void 
fermion_force_hisq_block( Real eps, Real *residues, 
			    su3_vector **xxx, int nterms, int veclength, 
			    int prec, ferm_links_t *fn, ks_action_paths *ap) 
{
  if(prec == 1)
    fermion_force_hisq_block_F( eps, residues, xxx, nterms, veclength,
				  fn, ap);
  else
    fermion_force_hisq_block_D( eps, residues, xxx, nterms, veclength,
				  fn, ap);

}

/**********************************************************************/
/*   Standard MILC interface for fermion force with multiple sources  */
/**********************************************************************/
void eo_fermion_force_multi( Real eps, Real *residues, su3_vector **xxx, 
			     int nterms, int prec, ferm_links_t *fn,
			     ks_action_paths *ap ) {

  int veclength;
#ifdef VECLENGTH
  veclength = VECLENGTH;
#else
  veclength = 4;
#endif
  /* Set default method to FNMAT. Set vector length for ASVEC, if
     requested */
  QOP_opt_t qop_ff_opt[2] = {
    {.tag = "fnmat_src_min",.value=4},
    {.tag = "veclength",.value=veclength}
  };

  switch(KS_MULTIFF){
  case ASVEC:
    qop_ff_opt[0].value = nterms + 1;  /* set high threshold for FNMAT */
    if(QOP_hisq_force_set_opts(qop_ff_opt, 2) != QOP_SUCCESS)
      node0_printf("eo_fermion_force_multi: error setting QOP options\n");
    fermion_force_hisq_block( eps, residues, xxx, nterms, 
				veclength, prec, fn, ap );
    break;
  default:  /* FNMAT */
    qop_ff_opt[0].value = 4; /* set sensible threshold for FNMAT */
    if(QOP_hisq_force_set_opts(qop_ff_opt, 2) != QOP_SUCCESS)
      node0_printf("eo_fermion_force_multi: error setting QOP options\n");
    fermion_force_hisq_multi( eps, residues, xxx, 
				nterms, prec, fn, ap );
  }
}

/**********************************************************************/
/*   Accessor for string describing the option                        */
/**********************************************************************/
const char 
*ks_multiff_opt_chr( void )
{
  switch(KS_MULTIFF){
  case ASVEC:
    return "ASVEC";
    break;
  default:
    return "FNMAT";
  }
  return NULL;
}

